/* This shell sort is basically an insertion sort which uses
   a gap of h instead of a gap of 1. In fact, this version
   of shell sort was created by a straight forward hack
   on the code for the insertion sort (I later continued
   hacking to get better optimization).
   
   Based on information in "Algorithms", by Robert Sedgewick,
   2nd ed., Addison-Wesley, 1988. */

#include <stdlib.h>
#include <string.h>
#include "sort_private.h"

static void shsort1(void *base, size_t nmemb, size_t size,
		    int (*compar)(const void *, const void *), void *v);

static void shsort2(void *base, size_t nmemb, size_t size,
		    int (*compar)(const void *, const void *));

/* Call shsort just like you would the standard library routine qsort.
   First, shsort tries to allocate a temporary variable; if this is
   successful, then it uses the function shsort1, which performs about 1/3
   of the data moves of shsort2. Otherwise, it uses shsort2, which uses
   the swap function.

   The code for both shsort1 and shsort2 has been optimized to use
   pointers, instead of indirectly accessing the elements of the array
   using a multiplication. This, unfortunately, resulted in slightly more
   obscure code. */

void shsort(void *base, size_t n, size_t size,
	    int (*compar)(const void *, const void *))
{
  void *v;
  
  v = SORT_MALLOC(size);
  if (v) {
    shsort1(base, n, size, compar, v);
    SORT_FREE(v);
  }
  else
    shsort2(base, n, size, compar);
}

static void shsort1(void *base, size_t n, size_t size,
		    int (*compar)(const void *, const void *), void *v)
{
  size_t h;     /* Distance between each two items. */
  size_t gap;   /* Gap in bytes between each two items (equal to h*size) */
  void *i;      /* Index for middle loop. */
  char *j;	/* Index for inner loop. */
  void *first;	/* Second item in first file; serves as a sentinel for j. */
  void *last;	/* Last item in array; serves as a sentinel for i. */
  
  last = get(n);
  for (h = 0; h < n; h = 3 * h + 1)
    ;
  for (h /= 3; h > 0; h /= 3) {
    gap = h * size;
    for (i = first = get(h); i < last; i = succ(i)) {
      COPY(v, i, size);
      for (j = i; (void *) j >= first && COMPARE(j - gap, v) > 0; j -= gap)
	COPY(j, j - gap, size);
      COPY(j, v, size);
    }
  }
}

static void shsort2(void *base, size_t n, size_t size,
		    int (*compar)(const void *, const void *))
{
  size_t h;	/* Distance between each two items. */
  size_t gap;	/* Gap in bytes between each two items (equal to h*size) */
  void *i;	/* Index for middle loop. */
  char *j;	/* Index for inner loop. */
  void *first;	/* Second item in first file; serves as a sentinel for j. */
  void *last;	/* Last item in array; serves as a sentinel for i. */
  
  last = get(n);
  for (h = 0; h < n; h = 3 * h + 1)
    ;
  for (h /= 3; h > 0; h /= 3) {
    gap = h * size;
    for (i = first = get(h); i < last; i = succ(i))
      for (j = i; (void *) j >= first && COMPARE(j - gap, j) > 0; j -= gap)
	SWAP(j, j - gap, size);
  }
}
