/* Copyright (c) 1992 Ari I. Halberstadt */

/* definitions private to the sort library */

#ifndef _H_SORT_PRIVATE
#define _H_SORT_PRIVATE

#include "sortlib.h"

/* get the i'th item from the array (base and size are local variables) */
#define get(i) ((void*)((char*)base + size * (i)))

/* return predecessor and succesor of pointer p (size is a local variable) */
#define pred(p) ((void*)((char*)(p) - size))
#define succ(p) ((void*)((char*)(p) + size))

/* memory allocation macros */
#if SORT_NOMEM
  #define SORT_MALLOC(n)  (NULL)
  #define SORT_FREE(p)	  ((void) 0)
#else /* SORT_NOMEM */
  #define SORT_MALLOC(n)  malloc(n)
  #define SORT_FREE(p)	  free(p)
#endif /* SORT_NOMEM */

/* The execution time of the sorts might be significantly improved by
   replacing calls to the comparison, copying, and swap functions
   with in-line code or with code optimized for the specific
   data. The greatest improvement will be noticed if you replace
   the comparison function with in-line code. For instance,
   when sorting integers try using an expression such as
   
     (*(int*)(a) - *(int*)(b))
     
   Be careful that any macro you use does not evaluate any
   of its parameters more than once. Also, the COPY macro
   should be compatible with memcpy. */

/* Compare A and B. Must not evaluate either of its parameters more
   than once. The function "compar" is a local variable. */
#ifndef COMPARE
  #define COMPARE(a, b)	compar(a, b)
#endif

/* Copy the first N bytes of B to A. Must not evaluate either of its
   parameters more than once. */ 
#define COPY(a, b, n)	memcpy(a, b, n)

/* Swap the first N bytes of A and B. Using a macro is usually
   faster than calling a function. May evaluate its parameters
   more than once. */
#define SWAP(a, b, n)\
{ \
  register char *_a = a, *_b = b, _tmp; \
  register size_t _n = n; \
  \
  while (_n-- > 0) { \
    _tmp = *_a; \
    *_a++ = *_b; \
    *_b++ = _tmp; \
  } \
}

/* private insertion sort for qksort */
void _insort(void *base, size_t nmemb, register size_t size,
  register int (*compar)(const void *, const void *), size_t n);
  
#endif
